%{
/*
   Hoarescope - Use Hoare logic for program correctness proofs
   Copyright (C) 2004 Markus Triska

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA

   Contact me at triska@gmx.at
*/

#include <stdlib.h>
#include "node.h"
extern int yyerror(char *);
extern int yylex();
extern nodeType *opr(int, int , ...);
extern nodeType *id(char *);
extern nodeType *con(int );
extern int prove_program(nodeType *, nodeType *, nodeType *);
extern void register_alias(nodeType *, char *);
%}
%union {
	int tok_ivalue;
	char *tok_string;
	nodeType *nPtr;
}

%token <tok_ivalue> NUMBER
%token <tok_string> LOWCASESTRING UPCASESTRING
%token BEGINTOK ENDTOK IF THEN ELSETOK GE LE EQ NE WHILE DO ASSIGN
%token EOL IMPLIES AND OR NOT SUBST WITH COMPOUND EXISTS FORALL
%token WHERE IS
%left AND OR IMPLIES
%left GE EQ NE LE '<' '>'
%left '+' '-'
%left '*' '/'
%nonassoc NOTPREC
%type <nPtr> stmt stmt_list expr condition variable
%%

commands: /* empty */
	| commands EOL
	| commands condition '{' stmt_list '}' condition EOL {
		prove_program($2, $4, $6); }
	| commands condition '{' stmt_list '}' condition WHERE alias_list EOL {
		prove_program($2, $4, $6); }
       ;

stmt:  IF condition THEN stmt_list ELSETOK stmt_list {
		$$ = opr(IF, 3, $2, $4, $6); }
	| WHILE condition DO stmt_list {
		$$ = opr(WHILE, 2, $2, $4); }
	| variable ASSIGN expr  { $$ = opr(ASSIGN, 2, $1, $3); }
      ;

stmt_list: stmt            { $$ = $1; }
	| BEGINTOK stmt_list ';' stmt_list ENDTOK
		{ $$ = opr(COMPOUND, 2, $2, $4); }
      ;

condition: UPCASESTRING    { $$ = id($1); free($1); }
	| EXISTS variable condition %prec NOTPREC { $$ = opr(EXISTS, 2, $2, $3); }
	| FORALL variable condition %prec NOTPREC { $$ = opr(FORALL, 2, $2, $3); }
	| expr '<' expr    { $$ = opr('<', 2, $1, $3); }
	| expr '>' expr    { $$ = opr('>', 2, $1, $3); }
	| expr EQ expr     { $$ = opr(EQ, 2, $1, $3); }
	| expr GE expr     { $$ = opr(GE, 2, $1, $3); }
	| expr NE expr     { $$ = opr(NE, 2, $1, $3); }
	| expr LE expr     { $$ = opr(LE, 2, $1, $3); }
	| condition IMPLIES condition { $$ = opr(IMPLIES, 2, $1, $3); }
	| condition AND condition     { $$ = opr(AND, 2, $1, $3); }
	| condition OR condition      { $$ = opr(OR, 2, $1, $3); }
	| NOT condition %prec NOTPREC { $$ = opr(NOT, 1, $2); }
	| '(' condition ')' { $$ = $2; }
	;


expr:   NUMBER           { $$ = con($1); }
	| variable         { $$ = $1; }
	| expr '+' expr    { $$ = opr('+', 2, $1, $3); }
	| expr '-' expr    { $$ = opr('-', 2, $1, $3); }
	| expr '*' expr    { $$ = opr('*', 2, $1, $3); }
	| expr '/' expr    { $$ = opr('/', 2, $1, $3); }
	| '(' expr ')'     { $$ = $2; }
	;

variable: LOWCASESTRING  { $$ = id($1); free ($1); }
        ;

alias_list: alias
	| alias_list ',' alias
	;

alias: condition IS UPCASESTRING {
		register_alias($1, $3);
		free($3);
	}
	;



